"""Simple program to test and visualize the ramp produced by the GainRamp effect."""
import numpy as np
import matplotlib.pyplot as plt
import asmu

def plot_gainramp():
    # create objects
    interface = asmu.Interface(samplerate=44100,
                               blocksize=1024,
                               analog_output_channels=[1],
                               no_audio_device = True)
    const = asmu.generator.Constant(interface, 1)
    gainramp = asmu.effect.GainRamp(interface, 0, 0.1, scale = "log")

    # establish connections
    const.output().connect(gainramp.input())
    gainramp.output().connect(interface.ioutput(ch=1))

    # inti plot
    fig, ax = plt.subplots()

    # setup vectors
    outdata = np.empty((interface.blocksize, 1), dtype=np.float32)
    x = np.linspace(0, 1, interface.blocksize)

    # set gain
    gainramp.set_gain(1.0)

    # run callback loop
    for i in range(12):
        interface.callback(None, outdata, None, None, None)
        ax.plot(x+i, outdata, label=f"frame_{i}")

    # setup plot and show
    ax.set(title="GainRamp for gain = 1, step = 0.1, scale = \"log\"", xlabel="frames", ylabel="gain")
    ax.grid()
    ax.legend()

    # Uncomment this if you dont want to save the plot
    # fig.savefig("docs/imgs/plot_gainramp.png", dpi=300)

    plt.show()

if __name__ == "__main__":
    plot_gainramp()



"""In this submodule you can find the Input and Output classes used in the "audio processors". 
You will rarely use them directly, but the special IInput and IOutput classes used by the interface, 
store a lot of important information you may want to access or modify."""
import contextlib
import logging
from typing import TYPE_CHECKING, Tuple, Optional, Union
import numpy as np

if TYPE_CHECKING:
    from .interface import Interface
    from .processor import Analyzer, Effect, Generator

logger = logging.getLogger(__name__)


class Input:
    def __init__(self, processor: Union["Effect", "Interface", "Analyzer"]):
        """Input base class.

        Args:
            processor: Reference to the corresponding Processor object.
        """
        self._processor = processor
        self._output: Optional["Output"] = None

    def __del__(self):
        if self._output is not None:
            self._output.disconnect(self)
        else:
            logger.info("No output to disconnect.")

    @property
    def connected_output(self) -> Optional["Output"]:
        return self._output
    @connected_output.setter
    def connected_output(self, value: Optional["Output"]):
        self._output = value
        # if an output was set or unset, update acore module of processor
        self._processor.update_acore()


class Output:
    def __init__(self, processor: Union["Generator", "Effect", "Interface"]):
        """Output base class.

        Args:
            processor: Reference to the corresponding Processor object.
        """
        self._processor = processor
        self._inputs: Tuple[Input] = ()

    def __del__(self):
        for inpu in self.inputs:
            self.disconnect(inpu)

    @property
    def inputs(self):
        return self._inputs
    
    @property
    def idx(self) -> int:
        return self._processor.outputs.index(self)
    
    @property
    def acore(self):
        return self._processor.acore
    
    def connect(self, input: "Input"):
        """Connect this output to the given Input.

        Args:
            input: The Input to connect this output to.

        Example: Connect devices
            ```python
            import asmu

            interface = asmu.Interface()
            sine = asmu.generator.Sine(interface, 1000)
            gain = asmu.effect.Gain(interface, 0.5)

            sine.output().connect(gain.input())
            gain.output().connect(interface.ioutput(ch = 2))
            ```
        """
        if input not in self._inputs:
            self._inputs += (input, )
            input.connected_output = self
            self._processor.update_acore()
        else:
            logger.info("Input is already connected to output.")

    def disconnect(self, input: "Input"):
        """Disconnect this output from the given Input.

        Args:
            input: The Input to disconnect this output from.
        """
        if input.connected_output is self and input in self._inputs:
            self._inputs = tuple(inp for inp in self._inputs if inp != self)
            input.connected_output = None
            self._processor.update_acore()
        else:
            logger.info("Trying to disconnect an input that is not connected to this output")


class IInput(Output):
    def __init__(self,
                 interface: "Interface",
                 channel: int, 
                 reference: Optional[bool] = None,
                 name: Optional[str] = None,
                 gain: Optional[float] = None,
                 latency: Optional[int] = None,
                 color: Optional[str] = None,
                 cPa: Optional[float] = None,
                 fPa: Optional[float] = None,
                 cV: Optional[float] = None,
                 fV: Optional[float] = None,
                 cFR: Optional[np.ndarray] = None,
                 fFR: Optional[np.ndarray] = None):
        """A special type of Output class used for the analog interface inputs.
        It stores various settings and options.

        Args:
            interface: Reference to an Interface instance.
            channel: Channel number on the interface.
            reference: Flag if the channel is used as reference for computation/calibration.
            name: Trivial name.
            gain: The gain setting of the input.
            latency: Individual IO latency (relative to Interface's system latency).
            color: A color used for plotting.
            cPa: Pressure calibration factor in Pascal.
            fPa: Frequency used for pressure calibration.
            cV: Voltage calibration factor in Volts.
            fV: Frequency used for voltage calibration.
            cFR: Frequency response calibration vector.
            fFR: Corresponding frequency vector.
        """
        self._interface = interface
        self.channel = channel
        self.reference = reference
        self.name = name
        self.gain = gain
        self.latency = latency
        self.color = color
        self.cPa = cPa
        self.fPa = fPa
        self.cV = cV
        self.fV = fV
        self.cFR = cFR
        self.fFR = fFR
        super().__init__(interface)

    def serialize(self) -> dict:
        data = {}
        data["channel"] = self.channel
        if self.reference is not None: data["reference"] = bool(self.reference)
        if self.name is not None: data["name"] = self.name
        if self.gain is not None: data["gain"] = float(self.gain)
        if self.latency is not None: data["latency"] = int(self.latency)
        if self.color is not None: data["color"] = self.color
        if self.cPa is not None: data["cPa"] = float(self.cPa)
        if self.fPa is not None: data["fPa"] = float(self.fPa)
        if self.cV is not None: data["cV"] = float(self.cV)
        if self.fV is not None: data["fV"] = float(self.fV)
        if self.cFR is not None:
            path = self._interface.path.with_suffix(f"/in_ch{self.channel:3.0f}_cFR.npy")
            np.save(path, self.cFR)
            data["cFR"] = path
        if self.fFR is not None:
            path = self._interface.path.with_suffix(f"/in_ch{self.channel:3.0f}_fFR.npy")
            np.save(path, self.fFR)
            data["fFR"] = path
        return data

    def deserialize(self, data: dict) -> None:
        self.channel = int(data["channel"])
        with contextlib.suppress(KeyError): self.reference = bool(data["reference"])
        with contextlib.suppress(KeyError): self.name = data["name"]
        with contextlib.suppress(KeyError): self.gain = float(data["gain"])
        with contextlib.suppress(KeyError): self.latency = int(data["latency"])
        with contextlib.suppress(KeyError): self.color = data["color"]
        with contextlib.suppress(KeyError): self.cPa = float(data["cPa"])
        with contextlib.suppress(KeyError): self.fPa = float(data["fPa"])
        with contextlib.suppress(KeyError): self.cV = float(data["cV"])
        with contextlib.suppress(KeyError): self.fV = float(data["fV"])
        with contextlib.suppress(KeyError): self.cFR = np.load(data["cFR"])
        with contextlib.suppress(KeyError): self.fFR = np.load(data["fFR"])


class IOutput(Input):
    def __init__(self,
                 interface: "Interface",
                 channel: int, 
                 reference: Optional[bool] = None,
                 name: Optional[str] = None,
                 gain: Optional[float] = None,
                 latency: Optional[int] = None,
                 color: Optional[str] = None,
                 cPa: Optional[float] = None,
                 fPa: Optional[float] = None,
                 cV: Optional[float] = None,
                 fV: Optional[float] = None,
                 cFR: Optional[np.ndarray] = None,
                 fFR: Optional[np.ndarray] = None):
        """A special type of Input class used for the analog interface outputs.
        It stores various settings and options.

        Args:
            interface: Reference to an Interface instance.
            channel: Channel number on the interface.
            reference: Flag if the channel is used as reference for computation/calibration.
            name: Trivial name.
            gain: The gain setting of the output.
            latency: Individual IO latency (relative to Interface's system latency).
            color: A color used for plotting.
            cPa: Pressure calibration factor in Pascal.
            fPa: Frequency used for pressure calibration.
            cV: Voltage calibration factor in Volts.
            fV: Frequency used for voltage calibration.
            cFR: Frequency response calibration vector.
            fFR: Corresponding frequency vector.
        """
        self._interface = interface
        self.channel = channel
        self.reference = reference
        self.name = name
        self.gain = gain
        self.latency = latency
        self.color = color
        self.cPa = cPa
        self.fPa = fPa
        self.cV = cV
        self.fV = fV
        self.cFR = cFR
        self.fFR = fFR
        super().__init__(interface)

    def serialize(self) -> dict:
        data = {}
        data["channel"] = self.channel
        if self.reference is not None: data["reference"] = self.reference
        if self.name is not None: data["name"] = self.name
        if self.gain is not None: data["gain"] = float(self.gain)
        if self.latency is not None: data["latency"] = int(self.latency)
        if self.color is not None: data["color"] = self.color
        if self.cPa is not None: data["cPa"] = float(self.cPa)
        if self.fPa is not None: data["fPa"] = float(self.fPa)
        if self.cV is not None: data["cV"] = float(self.cV)
        if self.fV is not None: data["fV"] = float(self.fV)
        if self.cFR is not None:
            path = self._interface.path.with_suffix(f"/out_ch{self.channel:3.0f}_cFR.npy")
            np.save(path, self.cFR)
            data["cFR"] = path
        if self.fFR is not None:
            path = self._interface.path.with_suffix(f"/out_ch{self.channel:3.0f}_fFR.npy")
            np.save(path, self.fFR)
            data["fFR"] = path
        return data
    
    def deserialize(self, data: dict, hashmap: dict = {}) -> None:
        self.channel = data["channel"]
        with contextlib.suppress(KeyError): self.reference = data["reference"]
        with contextlib.suppress(KeyError): self.name = data["name"]
        with contextlib.suppress(KeyError): self.gain = float(data["gain"])
        with contextlib.suppress(KeyError): self.latency = int(data["latency"])
        with contextlib.suppress(KeyError): self.color = data["color"]
        with contextlib.suppress(KeyError): self.cPa = float(data["cPa"])
        with contextlib.suppress(KeyError): self.fPa = float(data["fPa"])
        with contextlib.suppress(KeyError): self.cV = float(data["cV"])
        with contextlib.suppress(KeyError): self.fV = float(data["fV"])
        with contextlib.suppress(KeyError): self.cFR = np.load(data["cFR"])
        with contextlib.suppress(KeyError): self.fFR = np.load(data["fFR"])



from typing import TYPE_CHECKING, Annotated
import numpy as np
import scipy as sp

if TYPE_CHECKING:
    Data = Annotated[np.ndarray, "Blocksize x Channels"]
    RefData = Annotated[np.ndarray, "Blocksize"]
    Peaks = Annotated[np.ndarray, "Channels"]

def get_corrs_sampleshifts(indata: "Data", refdata: "RefData", burstLength: int) -> tuple["Data", "Peaks"]:
    """This function calculates the crosscorrelation between all channels of the indata relative to refdata.
    Additionally the maximum peak is exported. 

    Args:
        indata: Input data
        refdata: Reference data
        burstLength: The length of the burst in samples, to better fit the peak finding algorithm

    Returns:
        The full cross correlations in the format (Blocksize x Channel) and a vector containing the peak positions for each channel
    """
    corrs = []
    sampleshifts = []
    n = int(np.size(refdata))
    for cidx in range(np.ma.size(indata, axis=1)):
        v = indata[:, cidx]
        corr = sp.signal.correlate(v, refdata)
        corr /= np.max(corr)
        peaks, _ = sp.signal.find_peaks(corr, prominence=0.8, distance=burstLength, height=0.5)

        # shift correlation correctly
        corrs.append(corr[n-1:])
        sampleshifts.append(peaks[0]-(n-1))
    return np.array(corrs).T, np.array(sampleshifts)


"""conftest.py
Custom PyTest configuration.
"""

def pytest_addoption(parser):
    parser.addoption("--buffer", action="store_true", help="test all possible buffer settings")

def pytest_generate_tests(metafunc):
    # test both in_buffer settings (where possible)
    if "in_buffer" in metafunc.fixturenames:
        if metafunc.config.getoption("buffer"):
            in_buffer = [True, False]
        else:
            in_buffer = [True]
        metafunc.parametrize("in_buffer", in_buffer)
    # test both out_buffer settings (where possible)
    if "out_buffer" in metafunc.fixturenames:
        if metafunc.config.getoption("buffer"):
            out_buffer = [False, True]
        else:
            out_buffer = [False]
        metafunc.parametrize("out_buffer", out_buffer)



"""Welcome to the API of the asmu module!

To get started, navigate through the submodules on the right.

!!! info
    The [asmu.io][] is rarely used directly, but can be usefull to access the paramteres of audio interface IO.
    [asmu.processor][] and [asmu.acore][] are used for custum devices, that should probabpy be implemented in the package source code directly.
"""
# enable sounddevice ASIO
import os
os.environ["SD_ENABLE_ASIO"] = "1"

from .afile import AFile
from .asetup import ASetup
from .interface import Interface
from . import io
from . import generator
from . import effect
from . import analyzer

__all__ = ["Interface", "ASetup", "AFile"]

# enable logging
import logging
logging.getLogger("asmu").addHandler(logging.NullHandler())

def query_devices(device = None, kind = None):
    import sounddevice as sd
    return sd.query_devices(device, kind)



"""This example can be used to calibrate an Interface IInput and IOutput channel for voltage."""
import asmu

def calibrate_iinput_cV(interface: "asmu.Interface", in_ch: int):
    calcV = asmu.analyzer.CalIInput(interface, 1, "V", averages=100)
    interface.iinput(ch = in_ch).connect(calcV.input())

    stream = interface.start()
    while not calcV.finished():
        pass
    stream.stop()
    stream.close()
    print(calcV.evaluate())
    print(f"cV = {interface.iinput(ch = in_ch).cV}")
    print(f"fV = {interface.iinput(ch = in_ch).fV}")

def calibrate_ioutput_cV(interface: "asmu.Interface", in_ch: int, out_ch: int):
    outgain = 0.1
    sine = asmu.generator.Sine(interface, 1000)
    gain = asmu.effect.Gain(interface, outgain)
    calcV = asmu.analyzer.CalIOutput(interface, outgain, "V", interface.ioutput(ch = out_ch), averages=100)
    sine.output().connect(gain.input())
    gain.output().connect(interface.ioutput(ch = out_ch))
    interface.iinput(ch = in_ch).connect(calcV.input())

    stream = interface.start()
    while not calcV.finished():
        pass
    stream.stop()
    stream.close()
    print(calcV.evaluate())
    print(f"cV = {interface.ioutput(ch = out_ch).cV}")
    print(f"fV = {interface.ioutput(ch = out_ch).fV}")

if __name__ == "__main__":
    in_ch = 3
    out_ch = 3
    interface = asmu.Interface(device = "ASIO MADIface USB",
                               analog_input_channels=[in_ch],
                               analog_output_channels=[out_ch],
                               blocksize=8192,
                               samplerate=192000)

    print(f"Connect a 1Vp sine generator to input channel {in_ch}.")
    input("\tPress ENTER to start.")
    calibrate_iinput_cV(interface, in_ch)

    print(f"Disconnect the funtion generator and connect the input channel {in_ch} to the output channel {out_ch}?")
    input("\tPress ENTER to start.")
    calibrate_ioutput_cV(interface, in_ch, out_ch)

    if input("Start generator? (y|n)") == "y":
        # verify if everything worked correctly
        sine = asmu.generator.Sine(interface, 1000)
        Vp = 0.5 # set desired peak amplitude
        gain = asmu.effect.Gain(interface, Vp/interface.ioutput().cV)
        sine.output().connect(gain.input())
        gain.output().connect(interface.ioutput(ch = out_ch))
        print("Starting sine generator...")
        stream = interface.start()
        print(f"You now should measure a {Vp:.2f}Vp sine wave on the output channel {interface.ioutput().channel}.")
        input("\tPress ENTER to stop.")
        stream.stop()
        stream.close()



"""File that stores useful commands and a small test code for profiling."""
import numpy as np

# check for line_profiler or memory_profiler in the local scope, both
# are injected by their respective tools or they're absent
# if these tools aren't being used (in which case we need to substitute
# a dummy @profile decorator)
if 'line_profiler' not in dir() and 'profile' not in dir():
    def profile(func):
        def inner(*args, **kwargs):
            return func(*args, **kwargs)
        return inner

@profile
def main():
    a = np.ones((100000, 2))
    b = np.empty(100000)

    np.copyto(b, a[:, 1])
    b[:] = a[:, 1]

if __name__ == "__main__":
    main()
    
# kernprof -l -v ./profiling.py
# python -m memory_profiler ./profiling.py  


import serial
import numpy as np

# type "mode" in shell to find port and baudrate

class SparkCalibrator():
    def __init__(self, port: str, baud: int):
        self.port = port
        self.baud = baud

    def fire(self):
        try: 
            with serial.Serial(self.port, self.baud, parity=serial.PARITY_EVEN, stopbits=serial.STOPBITS_ONE, timeout=1) as s:
                s.flush()
                s.write("F\r".encode())
        except serial.SerialException:
            Exception(f"SparkCalibrator - could not open port {self.port}")


class BME280():
    def __init__(self, port: str, baud: int):
        self.port = port
        self.baud = baud

    def read(self):
        try:
            # open serial port
            with serial.Serial(self.port, self.baud, timeout=1) as s:
                # clear buffer
                s.reset_input_buffer()

                # block until read one line
                line = False
                while not line:
                    line = s.readline()

                # parse recieved line
                str_line = line.decode().rstrip()
                t, h, p = tuple(float(s) for s in str_line.strip("()").split(","))

                return (t, h/100, p)

        except serial.SerialException:
            print(f"BME280 - could not open port {self.port}")


def get_c0_cramer(t, h, p=98700):
    """Owen Cramer. “The variation of the specific heat ratio and the speed of sound in air with
    temperature, pressure, humidity, and CO2 concentration.” In: The Journal of the Acoustical
    Society of America 93.5 (1993), pp. 2510–2516. doi: 10 . 1121 / 1 . 405827. url: https :
    //doi.org/10.1121/1.405827."""

    a = [331.5024, 0.603055, -0.000528, 51.471935, 0.1495874,
        -0.000782, -1.82e-7, 3.73e-8, -2.93e-10, -85.20931,
        -0.228525, 5.91e-5, -2.835149, -2.15e-13, 29.179762, 0.000486]

    # thermodynamic temperature 
    T = t + 273.15 
    # enhancement  factor  for  moist  air 
    f = 1.0062 + 3.14e-8 * p + 5.6e-7 * t**2
    # saturation vapor pressure of water vapor in air
    psv = np.exp(1.2811805e-5 * T**2 - 1.9509874e-2 * T + 34.04926034 - 6.3536311e3 / T)

    xw = h*f*psv/p
    xc = 0.0004147

    return a[0] + a[1]*t + a[2]*t**2 + (a[3] + a[4]*t + a[5]*t**2)*xw \
        + (a[6] + a[7]*t + a[8]*t**2)*p + (a[9] + a[10]*t + a[11]*t**2)*xc \
        + a[12]*xw**2 + a[13]*p**2 + a[14]*xc**2 + a[15]*xw*p*xc


def get_c0_wong(t, h, p=None):
    """George S. K. Wong. “Characteristic impedance of humid air”. In: The Journal of the Acoustical
    Society of America 80.4 (Oct. 1, 1986), pp. 1203–1204. issn: 0001-4966, 1520-8524. doi: 10.1121/
    1.394468. url: https://pubs.aip.org/jasa/article/80/4/1203/770989/Characteristic-
    impedance-of-humid (visited on 10/03/2023)."""
    T0 = 273.15
    ch_cref = 1 + h*(9.66e-4 + 7.22e-5*t + 1.8e-6*t**2 + 7.2e-8*t**3 + 6.5e-11*t**4)
    return 331.29*ch_cref*((T0+t)/T0)**0.5

def get_Z0_wong(t, h, p):
    """George S. K. Wong. “Characteristic impedance of humid air”. en. In: The Journal of the Acoustical
    Society of America 80.4 (Oct. 1986), pp. 1203–1204. issn: 0001-4966, 1520-8524. doi: 10.1121/
    1.394468. url: https://pubs.aip.org/jasa/article/80/4/1203/770989/Characteristic-
    impedance-of-humid."""
    Zh_Zref = 1 - h*(1.3238e-3 + 1.02404e-4*t + 2.0624e-6*t**2 + 1.11e-7*t**3)
    return 428.11*(p)/101325*Zh_Zref*np.sqrt((273.15)/(273.15+t))


"""In this submodule you can find all analyzers, so "audio processors" with one or multiple inputs."""
import threading
from typing import TYPE_CHECKING, Literal, Union, Tuple
import numpy as np
from .acore import AAnalyzer
from .processor import Analyzer
from .io import Input

if TYPE_CHECKING:
    from .io import IInput, IOutput
    from .afile import AFile
    from .interface import Interface
    from .types import AData, FFTData


class Recorder(Analyzer):
    def __init__(self, interface: "Interface", afile: "AFile") -> None:
        """The Recorder class analyzer is used to record audio to a given file.
        It is a multi input analyzer, with the input count extracted from the given AFile.
        Input-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            afile: Reference to an AFile instance.

        Raises:
            ValueError: The given AFile was not opened.
        """
        # check afile is open and reset
        self._afile = afile
        if afile.closed:
            raise ValueError("The given AFile was not opened.")
        afile.flush()
        afile.seek(0)

        arecorder = self._ARecorder(afile, interface.blocksize, interface.start_frame)
        super().__init__(aanalyzer = arecorder,
                         interface = interface,
                         inputs = tuple(Input(self) for i in range(afile.channels)),
                         in_update = False)

    class _ARecorder(AAnalyzer):
        def __init__(self, afile: "AFile", blocksize, start_frame) -> None:
            self._afile = afile
            super().__init__(in_buffer = True,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _process(self) -> None:
            self._afile.write(self._in_buf)


class _CalIO(AAnalyzer):
        def __init__(self, averages: int, blocksize: int, start_frame: int) -> None:
            self.finished_event = threading.Event()

            # result averaging
            self._avgs = averages + 3 # we dont consider the first three measurements
            self.counter = 0
            self.us = np.zeros(self._avgs, dtype=np.float32)
            self.ks = np.zeros(self._avgs, dtype=np.float32)

            # arrays needed for the array processing
            self._fft = np.empty(int(blocksize/2+1), dtype=np.complex128)
            self._spectrum = np.empty(int(blocksize/2+1), dtype=np.float32)
            self._window = np.hanning(blocksize)
            super().__init__(in_buffer = True,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _process(self) -> None:
            if self.finished_event.is_set():
                return
            if self.counter < self._avgs:
                self.physical_fft(self._in_buf[:, 0], self._fft, window = self._window)
                np.abs(self._fft, out = self._spectrum)

                kmax = np.argmax(self._spectrum)
                self.ks[self.counter] = kmax
                self.us[self.counter] = self._spectrum[kmax]
                self.counter += 1
            elif not self.finished_event.is_set():
                self.finished_event.set()

        def physical_fft(self, indata: "AData", outdata: "FFTData", window: "AData" = 1) -> None:
            np.fft.rfft(indata*window, norm = "forward", out = outdata)
            outdata[:] *= 2/np.mean(window)


class CalIInput(Analyzer):
    def __init__(self, interface: "Interface", value: float, unit: Literal["V", "Pa", "SPL"], gain: float = 0, averages: int = 100) -> None:
        """The CalcIInput class analyzer is used to calibrate the connected interface IInput.
        It is a single input analyzer, therefore input-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            value: The value of the signal used for calibration.
            unit: The unit of the value given

                - `"V"`   : Peak amplitude of the sinusoidal signal in Volt.
                - `"Pa"`  : Peak amplitude of the sinusoidal signal in Pascal.
                - `"SPL"` : Sound Pressure Level (RMS pressure in Dezibel).

            gain: Gain setting of the interface. This is not used for the calculation, but stored in the IInput.
            averages: How many averages should be calcualted.
        """
        self._value = value
        self._unit = unit
        self._gain = gain

        self._calio = _CalIO(averages, interface.blocksize, interface.start_frame)
        super().__init__(aanalyzer = self._calio,
                         interface = interface,
                         inputs = (Input(self), ),
                         in_update = False)

    def evaluate(self, save: bool = True, iinput: "IInput" = None) -> Union[Tuple[float, float], False]:
        """If the measurement is finished, this evaluates the result and returns True.
        If the measurement is still running, False is returned.

        Args:
            save: Decides if the results should be daved to the connected or given IInput.
            iinput: The IInput to save the calibration to.

        Returns:
            (Frequency, Calibration Factor), when everything was successful. `False`, when the measurement is not done yet.

        Raises:
            ValueError: Given Unit is unknown.
        """
        if not self.finished(block = False):
            return False
        # calculate peak amplitude
        if self._unit == "V" or self._unit == "Pa":
            peak = self._value
        elif self._unit == "SPL":
            peak = 2e-5 * 10 ** (self._value / 20) * np.sqrt(2)
        else:
            raise ValueError("Given Unit is unknown.")
        # calculate calibration factor and frquency
        c = peak/np.mean(self._calio.us[2:])
        fs = np.fft.rfftfreq(self._interface.blocksize, 1/self._interface.samplerate)
        f = np.interp(np.mean(self._calio.ks[2:]), np.arange(fs.size), fs)
        # get iinput if given
        if iinput is None:
            iinput = self._inputs[0].connected_output
        # write to IInput
        if save:
            iinput.gain = self._gain
            if self._unit == "V":
                iinput.cV = c
                iinput.fV = f
            elif self._unit == "SPL" or self._unit == "Pa":
                iinput.cPa = c
                iinput.fPa = f
        return (f, c)
        
    def finished(self, block: bool = True, timeout: float = 1) -> bool:
        """This function can be used to wait for the calibration to be done.
        
        Args:
            block: Decides if the call of finished() should block.
            timeout: The timeout after which False is returned.

        Returns:
            `True`, when the calibration finished. `False` on timeout.
        """
        if block:
            return self._calio.finished_event.wait(timeout = timeout)
        else:
            return self._calio.finished_event.is_set()

    def restart(self) -> bool:
        """Restart the calibration, unset the finished() flag.
        If the audio stream is still running, it starts with the next measurement series right away.
        
        Returns:
            `True`, when successful. `False`, when the calibration isn't finished yet."""
        if not self.finished(block=False):
            return False
        self._calio.counter = 0
        self._calio.finished_event.clear()
        return True


class CalIOutput(Analyzer):
    def __init__(self, interface: "Interface", value: float, unit: Literal["V", "Pa"], ioutput: "IOutput", gain: float = 0, averages: int = 100) -> None:
        """The CalcIOutput class analyzer is used to calibrate the given IOutput.
        For this calibration procedure a Sine generator with an amplitude of value has to be connected to the given IOutput.
        This IOutput must be physically connected to the IInput, which needs to be pre-calibrated, 
        and this IInput must be connected to this analyzer.
        It is a single input analyzer, therefore input-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            value: The peak amplitude value of the signal used for calibration in arbitrary units.
            unit: The physical quantity to calibrate for.
            ioutput: The IOutput that is used to generate the signal and connected to the IInput.
            gain: Gain setting of the interface. This is not used for the calculation, but stored in the IOutput.
            averages: How many averages should be calcualted.
        """
        self._value = value
        self._unit = unit
        self._ioutput = ioutput
        self._gain = gain

        self._calio = _CalIO(averages, interface.blocksize, interface.start_frame)
        super().__init__(aanalyzer = self._calio,
                         interface = interface,
                         inputs = (Input(self), ),
                         in_update = False)

    def evaluate(self, save: bool = True, ioutput: "IOutput" = None, iinput: "IInput" = None) -> Union[Tuple[float, float], False]:
        """If the measurement is finished, this evaluates the result and returns True.
        If the measurement is still running, False is returned.

        Args:
            save: Decides if the results should be saved to the given IOutput.
            ioutput: The IOutput to save the calibration to.
            iinput: The IInput the analyzer is connected to. (Typically selected automatically)

        Returns:
            (Frequency, Calibration Factor), when everything was successful. `False`, when the measurement is not done yet.

        Raises:
            ValueError: Given Unit is unknown.
        """
        if not self.finished(block = False):
            return False
        # get iinput if not given
        if iinput is None:
            iinput = self._inputs[0].connected_output
        # get the measured value
        if self._unit == "V":
            peak = np.mean(self._calio.us[2:]) * iinput.cV
        elif self._unit == "Pa":
            peak = np.mean(self._calio.us[2:]) * iinput.cPa
        else:
            raise ValueError("Given Unit is unknown.")
        # calculate calibration factor and frquency
        c = peak/self._value
        fs = np.fft.rfftfreq(self._interface.blocksize, 1/self._interface.samplerate)
        f = np.interp(np.mean(self._calio.ks[2:]), np.arange(fs.size), fs)
        # get ioutput if not given
        if ioutput is None:
            ioutput = self._ioutput
        # write to IOutput
        if save:
            ioutput.gain = self._gain
            if self._unit == "V":
                ioutput.cV = c
                ioutput.fV = f
            elif self._unit == "Pa":
                ioutput.cPa = c
                ioutput.fPa = f
        return (f, c)

    def finished(self, block: bool = True, timeout: float = 1) -> bool:
        """This function can be used to wait for the calibration to be done.
        
        Args:
            block: Decides if the call of finished() should block.
            timeout: The timeout after which False is returned.

        Returns:
            `True` when the calibration finished. `False` on timeout.
        """
        if block:
            return self._calio.finished_event.wait(timeout = timeout)
        else:
            return self._calio.finished_event.is_set()

    def restart(self) -> bool:
        """Restart the calibration, unset the finished() flag.
        If the audio stream is still running, it starts with the next measurement series right away.
        
        Returns:
            `True` when successful. `False`, when the calibration isn't finished yet."""
        if not self.finished(block=False):
            return False
        self._calio.counter = 0
        self._calio.finished_event.clear()
        return True



"""test_gainramp.py
PyTest for the calibration analyzers CalIInput and CalIOutput.
"""
import numpy as np
import asmu

def test_caliinput(benchmark):
    Vp = 1
    outgain = 0.1
    # create objects
    interface = asmu.Interface(samplerate=44100,
                               blocksize=8192,
                               analog_output_channels=[1],
                               analog_input_channels=[1],
                               no_audio_device = True)
    sine = asmu.generator.Sine(interface, 1000)
    gain05 = asmu.effect.Gain(interface, 0.5)
    caliinput = asmu.analyzer.CalIInput(interface, Vp, "V")
    calioutput = asmu.analyzer.CalIOutput(interface, outgain, "V", interface.ioutput())

    # establish connections
    sine.output().connect(gain05.input())
    gain05.output().connect(caliinput.input())
    gain05.output().connect(calioutput.input())

    # create dummy array(s) for callback
    data = np.empty((interface.blocksize, 1), dtype = np.float32)
    # repeatedly call callback
    while not caliinput.finished(block = False):
        interface.callback(data, data, None, None, None)

    fV, cV = caliinput.evaluate(iinput=interface.iinput())
    # we used a sine with arbitrary amplitude 0.5 as our generated input signal, and a calibration value of 1Vp.
    # So to get from arbitrary to input voltage, cV should be 2.
    # The values are not perfect due to the limited window size.
    print(f"IInput: fV = {fV:.2f}Hz, cV = {cV:.2f}")
    assert abs(fV-1000) < 5, "Frequency deviates more than allowed range."
    assert abs(cV - Vp/0.5) < Vp/0.5*0.1, "Amplitude deviates more than allowed range."
    # check if calibration values were correctly written to IInput
    assert fV == interface.iinput().fV
    assert cV == interface.iinput().cV

    fV, cV = calioutput.evaluate(iinput = interface.iinput())
    # we theoretically generated a sine with arbitrary amplitude 0.1 and recieved a sine with a voltage amplitude of 1Vp.
    # So to get from voltage to the arbitrary amplitude used by the interface, 1/cV should be 0.1, so cV should be 10.
    # The values are not perfect due to the limited window size.
    print(f"IOutput: fV = {fV:.2f}Hz, cV = {cV:.2f}")
    assert abs(fV-1000) < 5, "Frequency deviates more than allowed range."
    assert abs(cV - Vp/outgain) < Vp/outgain*0.1, "Amplitude deviates more than allowed range."
    # check if calibration values were correctly written to IOutput
    assert fV == interface.ioutput().fV
    assert cV == interface.ioutput().cV

    # benchmark (calls callback very often)
    caliinput.restart()
    calioutput.restart()
    benchmark(interface.callback, data, data, None, None, None)


if __name__ == "__main__":
    test_caliinput(lambda x, *args: [x(*args) for _ in range(100)][0])



"""This file stores abreviation of complex custom types."""
import numpy as np
import numpy.typing as npt
from typing import Optional, Tuple, Union, Annotated
from .acore import AGenerator, AEffect, AAnalyzer, AInterface

# used for the in_as variable of the acore elements AEffect, AAnalyzer, AInerface
InAs = Optional[Tuple[Tuple[Optional[Union[AGenerator, AEffect, AInterface]], int], ...]]

# Used for the single channel data chunks transferred by all audio blocks
AData = Annotated[npt.NDArray[np.float32], "blocksize"]

# Used for FFt data chunks obtained by transforming a single channel data chunk
FFTData = Annotated[npt.NDArray[np.complex128], "int(blocksize/2+1)"]

# Used for the internal acore buffers
ABuf = Annotated[npt.NDArray[np.float32], "blocksize x channels"]

# Used for the internal acore buffers
AVector = Annotated[npt.NDArray[np.float32], "samples x channels"]

# all four acore elements
ACore = Union[AGenerator, AAnalyzer, AEffect, AInterface]









"""psense.py"""
import serial
from typing import List, Optional

class PSense():
    def __init__(self, port: str, baud: int):
        self.port = port
        self.baud = baud
        self.serial = None

    def __enter__(self) -> "PSense":
        try:
            self.serial = serial.Serial(self.port, self.baud, timeout=1)
        except serial.SerialException:
            print(f"PSense - could not open port {self.port}")
        return self
    
    def read_line(self) -> Optional[List[float]]:
        if self.serial is None:
            raise serial.SerialException("No connection opened openend.")

        # clear buffer
        self.serial.reset_input_buffer()

        # block until read one line
        line = False
        while not line:
            line = self.serial.readline()

        # parse recieved line
        str_line = line.decode("utf-8").strip()
        try:
            list = [float(s) for s in str_line.split(" ")]
        except ValueError:
            return None
        if len(list) == 16:
            return list
        else:
            return None
    
    def __exit__(self, *args) -> None:
        if self.serial is not None:
            self.serial.close()



"""In this submodule you can find all effects, so "audio processors" with one or multiple inputs and outputs."""
import queue
from typing import TYPE_CHECKING, Literal
import numpy as np
from .acore import AEffect
from .processor import Effect
from .io import Input, Output

if TYPE_CHECKING:
    from .interface import Interface
    from .types import AData


class Gain(Effect):
    def __init__(self, interface: "Interface", gain: float, in_buffer: bool = True, out_buffer: bool = False) -> None:
        """The Gain class effect is used to multiply a signal with the given gain.
        It is a multi input multi output effect, that applies the same gain to all connections.
        Input-update and output-update are supported.
        
        Args:
            interface: Reference to an Interface instance.
            gain: The gain.
            in_buffer: Flag that decides if inputs are buffered.
            out_buffer: Flag that decides if outputs are buffered.
        """
        self._gain = gain
        self._again = self._AGain(gain, in_buffer, out_buffer, interface.blocksize, interface.start_frame)
        super().__init__(aeffect = self._again,
                         interface = interface,
                         inputs = (Input(self), ),
                         outputs = (Output(self), ),
                         in_update = True,
                         out_update = True)

    class _AGain(AEffect):
        def __init__(self, gain: float, in_buffer: bool, out_buffer: bool, blocksize: int, start_frame: int) -> None:
            self._gain = gain
            super().__init__(in_buffer = in_buffer,
                             out_buffer = out_buffer,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            if self._in_buffer:
                outdata[:] = self._in_buf[:, ch] * self._gain
            else:
                if self._in_as[ch][0] is not None:
                    self._in_as[ch][0].upstream(outdata, self._in_as[ch][1], self._frame) # get outdata from upstream
                    outdata[:] *= self._gain # modify outdata

        def _inc(self) -> None:
            pass


class GainRamp(Effect):
    def __init__(self, interface: "Interface", gain: float, step: float, scale: Literal["lin", "log"] = "lin", in_buffer: bool = True, out_buffer: bool = False) -> None:
        """The GainRamp class effect is used to smoothly change between gains.
        The gain initially specified and the newly set values are linear setpoints. If scale is set to "log", this will be scaled accordingly.
        It is a multi input multi output effect, that applies the same gain to all connections.
        Input-update and output-update are supported.
        

        Args:
            interface: Reference to an Interface instance.
            gain: The initial gain.
            step: The desired linear gain change per frame.
            scale: Scaling of the given gain.
            in_buffer: Flag that decides if inputs are buffered.
            out_buffer: Flag that decides if outputs are buffered.
        """
        self._gain = gain
        self._againramp = self._AGainRamp(gain, step, scale, in_buffer, out_buffer, interface.blocksize, interface.start_frame)
        super().__init__(aeffect = self._againramp,
                         interface = interface,
                         inputs = (Input(self), ),
                         outputs = (Output(self), ),
                         in_update = True,
                         out_update = True)

    def set_gain(self, gain: float) -> None:
        """This is updated at the next upcoming frame.
        This is updated at the next upcoming frame.
        The function call blocks when called faster than the frames.
        
        Args:
            gain: The linear gain setpoint. If scale is set to "log", this will be scaled accordingly.
        """
        self._gain = gain
        self._againramp.gain_queue.put(self._gain)

    class _AGainRamp(AEffect):
        def __init__(self, gain: float, step: float, scale: Literal["lin", "log"], in_buffer: bool, out_buffer: bool, blocksize: int, start_frame: int) -> None:
            self.gain_queue = queue.Queue(maxsize=1)

            self._is_gain = gain
            self._set_gain = gain
            self._ramp = np.linspace(0, step, blocksize, endpoint=True, dtype=np.float32)
            self._scale = scale

            self._gainramp = np.ones_like(self._ramp, dtype=np.float32) * gain
            super().__init__(in_buffer = in_buffer,
                             out_buffer = out_buffer,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            if self._in_buffer:
                outdata[:] = self._in_buf[:, ch] * self._gainramp
            else:
                if self._in_as[ch][0] is not None:
                    self._in_as[ch][0].upstream(outdata, self._in_as[ch][1], self._frame) # get outdata from upstream
                    outdata[:] *= self._gainramp # modify outdata

        def _inc(self) -> None:
            # get new gain if available
            if not self.gain_queue.empty():
                self._set_gain = self.gain_queue.get()
            # check if we are still in ramp mode (with tolerance)
            if abs(self._set_gain - self._is_gain) > 0.001:
                if self._set_gain > self._is_gain:
                    # ramping downwards
                    self._gainramp = self._is_gain + self._ramp
                    self._gainramp[self._gainramp > self._set_gain] = self._set_gain
                else:
                    # ramping upwards
                    self._gainramp = self._is_gain - self._ramp
                    self._gainramp[self._gainramp < self._set_gain] = self._set_gain
                # set new gain
                self._is_gain = self._gainramp[-1]

                # scale the ramp
                if self._scale == "log":
                    self._log(self._gainramp)
            else:
                self._gainramp[:] = self._set_gain

        def _log(self, x: "AData") -> None:
            x += np.log10(1/9)
            np.power(10, x, out=x)
            x -= 1/9


class Sum(Effect):
    def __init__(self, interface: "Interface", out_buffer: int = False) -> None:
        """The Sum class effect is used to sum multiple inputs to a single output.
        Arithmetic averaging is used for summing.
        It is a multi input single output effect, therefore output-update is not supported.
        Input-update is supported.

        Args:
            interface: Reference to an Interface instance.
            out_buffer: Flag that decides if outputs are buffered.
        """
        self._asum = self._ASum(out_buffer, interface.blocksize, interface.start_frame)
        super().__init__(aeffect = self._asum,
                         interface = interface,
                         inputs = (Input(self), ),
                         outputs = (Output(self), ),
                         in_update = True,
                         out_update = False)

    class _ASum(AEffect):
        def __init__(self, out_buffer, blocksize, start_frame) -> None:
            super().__init__(in_buffer = True,
                             out_buffer = out_buffer,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            outdata[:] = np.mean(self._in_buf, axis=1)

        def _inc(self) -> None:
            pass



"""In this submodule you can find all generators, so "audio processors" with one or multiple outputs."""
import queue
import threading
from typing import TYPE_CHECKING
import numpy as np
from .acore import AGenerator
from .processor import Generator
from .io import Output

if TYPE_CHECKING:
    from .afile import AFile
    from .interface import Interface
    from .types import AData, AVector


class Player(Generator):
    def __init__(self, interface: "Interface", afile: "AFile", loop: bool = False) -> None:
        """The Player class generator is used to play audio of a given file.
        It is a multi output generator, with the output count extracted from the given AFile.
        Output-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            afile: Reference to an AFile instance.
            loop: Flag to enable looping.

        Raises:
            ValueError: The given AFile was not opened.
        """
        # check afile is open and reset
        self._afile = afile
        if afile.closed:
            raise ValueError("AFile was not opened - can not be used in Recorder.")
        afile.flush()
        afile.seek(0)

        self._loop = loop

        self._aplayer = self._APlayer(afile, loop, interface.blocksize, interface.start_frame)
        super().__init__(agenerator = self._aplayer,
                         interface = interface,
                         outputs = tuple(Output(self) for i in range(afile.channels)),
                         out_update = False)

    def finished(self, block: bool = True, timeout: float = 1) -> bool:
        """This function can be used to wait for the Player to finish playback.
        
        Args:
            block: Decides if the call of finished() should block.
            timeout: The timeout after which False is returned.

        Returns:
            `True`, when the Player finished. `False` on timeout.
        """
        if block:
            return self._aplayer.finished_event.wait(timeout = timeout)
        else:
            return self._aplayer.finished_event.is_set()

    def restart(self) -> bool:
        """Restart the Player, unset the finished() flag.
        
        Returns:
            `True`, when successful. `False`, when the Player isn't finished yet.
        """
        if not self.finished(block=False):
            return False
        self._afile.seek(0)
        self._aplayer.finished_event.clear()
        return True

    def looped(self, block: bool = True, timeout: float = 1) -> bool:
        """This function can be used to wait for the Player to loop.
        
        Args:
            block: Decides if the call of looped() should block.
            timeout: The timeout after which False is returned.

        Returns:
            `True`, when a loop occured. `False` on timeout.
        """
        looped = False
        if block:
            looped = self._aplayer.looped_event.wait(timeout = timeout)
        else:
            looped = self._aplayer.looped_event.is_set()
        if looped:
            self._aplayer.looped_event.clear()
        return looped

    def set_seek(self, position: int = 0) -> None:
        """Reset the Player to the given file position.
        This is updated at the next upcoming frame.
        The function call blocks when called faster than the frames.
        
        Args:
            position: Position to seek."""
        self._aplayer.seek_queue.put(position)

    class _APlayer(AGenerator):
        def __init__(self, afile: "AFile", loop: bool, blocksize: int, start_frame: int) -> None:
            self.finished_event = threading.Event()
            self.looped_event = threading.Event()
            self.seek_queue = queue.Queue()

            self._afile = afile
            self._loop = loop
            self._samples = afile.frames

            self._firstread = False
            super().__init__(out_buffer = True, # The _out_buf is used to write the file into on _inc()
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            if not self._firstread:
                self._inc()
                self._firstread = True

        def _inc(self) -> None:
            if self.finished_event.is_set():
                return
            if not self.seek_queue.empty():
                self._afile.seek(self.seek_queue.get())
            # calculate the rest to play
            rest = self._samples-self._afile.tell()
            if self._loop and rest < self._blocksize:
                # read rest (not full vlock)
                self._afile.read(rest, dtype="float32", always_2d=True, out=self._out_buf[:rest, :])
                self.looped_event.set() 
                self._afile.seek(0)
                self._afile.read(self._blocksize-rest, dtype="float32", always_2d=True, out=self._out_buf[rest:, :])
                return
            # if there are no samples left to play
            if rest == 0:
                self.finished_event.set() 
                return
            # load a block of samples in the buffer
            self._afile.read(self._blocksize, dtype="float32", always_2d=True, fill_value=0, out=self._out_buf)


class Sine(Generator):
    def __init__(self, interface: "Interface", frequency: float, phase: float = 0, out_buffer: bool = False) -> None:
        """The Sine class generator is used to craete a sine wave with given frequency and phase.
        It is a single output generator, therefore output-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            frequency: Sine frequency in Hertz.
            phase: Sine phase in radiant.
            out_buffer: Flag that decides if outputs are buffered.
        """
        self._frequency = frequency

        self._asine =  self._ASine(frequency, phase, interface.samplerate, out_buffer, interface.blocksize, interface.start_frame)
        super().__init__(agenerator = self._asine,
                         interface = interface,
                         outputs = (Output(self), ),
                         out_update = False)

    def set_frequency(self, frequency: float) -> None:
        """Change the frequency to the given value.
        This is updated at the next upcoming frame.
        The function call blocks when called faster than the frames.
        
        Args:
            frequency: The new set frequency in Hertz."""
        self._frequency = frequency
        self._asine.frequency_queue.put(frequency)

    class _ASine(AGenerator):
        def __init__(self, frequency, phase, samplerate, out_buffer, blocksize, start_frame) -> None:
            self.frequency_queue = queue.Queue(maxsize=1)

            self._frequency = frequency
            self._samplerate = samplerate
            self._phase = phase
            self._omega_per_block = 2*np.pi*frequency*blocksize/samplerate

            self._omegas = np.linspace(start_frame*self._omega_per_block, (start_frame+1)*self._omega_per_block, blocksize, endpoint=False, dtype=np.float32)
            super().__init__(out_buffer = out_buffer,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            np.sin(self._omegas+self._phase, out=outdata)

        def _inc(self) -> None:
            self._phase += self._omega_per_block
            self._phase %= 2*np.pi

            if not self.frequency_queue.empty():
                self._frequency = self.frequency_queue.get()
                self._omega_per_block = 2*np.pi*self._frequency*self._blocksize/self._samplerate
                self._omegas = np.linspace(0, self._omega_per_block, self._blocksize, endpoint=False, dtype=np.float32)


class SineBurst(Generator):
    def __init__(self, interface: "Interface", frequency: float, periods: int, phase: float = 0, out_buffer: bool = False) -> None:
        """The SineBurst class generator is used to craete a sine burst wave with given frequency and phase for given periods.
        It is a single output generator, therefore output-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            frequency: Sine frequency in Hertz.
            periods: The number of periods.
            phase: Sine phase in radiant.
            out_buffer: Flag that decides if outputs are buffered.
        """
        self._asine =  self._ASineBurst(frequency, periods, phase, interface.samplerate, out_buffer, interface.blocksize, interface.start_frame)
        super().__init__(agenerator = self._asine,
                         interface = interface,
                         outputs = (Output(self), ),
                         out_update = False)

    class _ASineBurst(AGenerator):
        def __init__(self, frequency, periods, phase, samplerate, out_buffer, blocksize, start_frame) -> None:
            self.frequency_queue = queue.Queue(maxsize=1)

            self._frequency = frequency
            self._samplerate = samplerate
            self._phase = phase
            self._maxang = phase + 2*np.pi*periods
            self._omega_per_block = 2*np.pi*frequency*blocksize/samplerate

            self._omegas = np.linspace(start_frame*self._omega_per_block, (start_frame+1)*self._omega_per_block, blocksize, endpoint=False, dtype=np.float32)
            super().__init__(out_buffer = out_buffer,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            if self._phase < self._maxang:
                np.sin(self._omegas+self._phase, out=outdata)
                outdata[self._omegas>=self._maxang] = 0

        def _inc(self) -> None:
            if self._phase > self._maxang:
                self._phase += self._omega_per_block


class Vector(Generator):
    def __init__(self, interface: "Interface", vector: "AVector") -> None:
        """The Vector class generator is used to craete a custom wave given by a numpy array.
        It is a multi output generator, with the output count extracted from axis 1 of the given array.
        Output-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            vector: Wave(s) to generate as numpy vector.
        """
        self._vector = vector.astype(np.float32, copy=True)

        self._avector =  self._AVector(self._vector, interface.blocksize, interface.start_frame)
        super().__init__(agenerator = self._avector,
                         interface = interface,
                         outputs = tuple(Output(self) for i in self._vector[0,:]),
                         out_update = False)
        
    def finished(self, block: bool = True, timeout: float = 1) -> bool:
        """This function can be used to wait for the Player to finish playback.
        
        Args:
            block: Decides if the call of finished() should block.
            timeout: The timeout after which False is returned.

        Returns:
            `True`, when the Player finished. `False` on timeout.
        """
        if block:
            return self._avector.finished_event.wait(timeout = timeout)
        else:
            return self._avector.finished_event.is_set()
        
    def restart(self) -> bool:
        """Restart the Player, unset the finished() flag.
        
        Returns:
            `True`, when successful. `False`, when the Player isn't finished yet.
        """
        if not self.finished(block=False):
            return False
        self._avector.pos = 0
        self._avector.finished_event.clear()
        return True

    class _AVector(AGenerator):
        def __init__(self, vector, blocksize, start_frame) -> None:
            self.finished_event = threading.Event()
            self._vector = vector
            self._samples = np.size(vector[:,0])
            self.pos = 0

            super().__init__(out_buffer = False,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            if self.finished_event.is_set():
                return
            rest = self._samples-self.pos
            if rest < self._blocksize:
                # read rest (not full vlock)
                outdata[:rest] = self._vector[:, ch]
                return
            outdata[:] = self._vector[:, ch]

        def _inc(self) -> None:
            if self.finished_event.is_set():
                return
            self.pos += self._blocksize
            # if there are no samples left to play
            if self._samples - self.pos <= 0:
                self.finished_event.set() 
                return


class Constant(Generator):
    def __init__(self, interface: "Interface", value: float) -> None:
        """The Constant class generator is used to generate a constant output value, typically used for testing.
        It is a single output generator, therefore output-update is not supported.

        Args:
            interface: Reference to an Interface instance.
            value: The constant output value
        """
        self._aconstant =  self._AConstant(value, interface.blocksize, interface.start_frame)
        super().__init__(agenerator = self._aconstant,
                         interface = interface,
                         outputs = (Output(self), ),
                         out_update = False)

    class _AConstant(AGenerator):
        def __init__(self, value, blocksize, start_frame) -> None:
            self._value = value
            super().__init__(out_buffer = False,
                             blocksize = blocksize,
                             start_frame = start_frame)

        def _mod(self, outdata: "AData", ch: int) -> None:
            outdata[:] = self._value

        def _inc(self) -> None:
            pass



import json
import pathlib
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .interface import Interface

logger = logging.getLogger(__name__)


class ASetup:
    def __init__(self, path: str, ) -> None:
        """The ASetup class handles .asmu JSON files, which store general Interface settings, IInput/IOutput configuration and calibration values. 

        Args:
            path: Path to .asmu file.
        """
        self.path = pathlib.Path(path)
        self.interface: "Interface" = None
        # add time/date and other info here
        now = datetime.now()
        self.date = now.strftime("%Y-%m-%dT%H:%M:%S%z") # ISO 8601


    def __del__(self):
        if self.interface is not None:
            self.interface.asetup = None

    def load(self):
        """Load setup data from given path.
        This is typically called by the interface, if you specify the `asetup` at initialization.
        When the `asetup` is specified at a later point, this method can be called manually to load the settings.

        Raises:
            ValueError: No associated Interface to load to.
        """
        if self.interface is None:
            raise ValueError("No associated Interface to load to.")
        with open(self.path, "r", encoding="utf-8") as asetup:
            self.deserialize(json.load(asetup))

    def save(self, path: Optional[str] = None):
        """Save setup data to given path.

        Args:
            path: When given, the setup is saved to this path instead. (Save Copy As)

        Raises:
            ValueError: No associated Interface to save from.
        """
        if self.interface is None:
            raise ValueError("No associated Interface to save from.")

        if path is not None:
            path = pathlib.Path(path)
        else:
            path = self.path

        with open(path, "w", encoding="utf-8") as asetup:
            asetup.write(json.dumps(self.serialize(), sort_keys=True, indent=4, separators=(',', ': ')))

    def serialize(self):
        data = self.interface.serialize()
        data["created"] = self.date
        return data

    def deserialize(self, data: dict):
        self.date = data["created"]
        self.interface.deserialize(data)



"""test_gainramp.py
PyTest for the GainRamp effect, with "log" scale and for all buffer settings.
"""

import numpy as np
import asmu

def test_gainramp(in_buffer, out_buffer, benchmark):
    step = 0.1
    # create objects
    interface = asmu.Interface(samplerate=44100,
                               blocksize=1024,
                               analog_output_channels=[1],
                               no_audio_device = True)
    const = asmu.generator.Constant(interface, 1)
    gainramp = asmu.effect.GainRamp(interface, 0, step, scale = "log", in_buffer = in_buffer, out_buffer = out_buffer)

    # establish connections
    const.output().connect(gainramp.input())
    gainramp.output().connect(interface.ioutput(ch=1))

    # setup vector for callback to write to
    outdata = np.empty((interface.blocksize, 1), dtype=np.float32)

    # call callback once (_inc() is not called)
    interface.callback(None, outdata, None, None, None)

    # set new gain
    set_gain = 0.9
    gainramp.set_gain(set_gain)

    # for the second callback call _inc() is called with the new gain
    interface.callback(None, outdata, None, None, None)

    # check if logarithmic step was performed correctly
    log_step = 10**(step + np.log10(1/9)) - 1/9
    assert abs(outdata[-1, 0] - log_step) < 1e-6

    # benchmark (calls callback very often)
    benchmark(interface.callback, None, outdata, None, None, None)

    # check if set_gain was reached
    assert abs(outdata[-1, 0] - set_gain) < 1e-6

if __name__ == "__main__":
    test_gainramp(True, False, lambda x, *args: [x(*args) for _ in range(100)][0])




"""The signals are handled by the ACore objects, which are optimized Python classes which are called in the callback of the audio interface.
They handle the audio buffers and call the other connected ACore objects. The execution time and memory usage of these functions is critical, always use [profiling](../development.md#profiling) when working on these base classes, or inherit from them for new processors. Higher memory usage of ACore functions can increase the thread switching time drastically; please akeep that in mind.

!!! quote "General philosophy"
    ACore objects are fast audio manipulation classes, that should never dynamically allocate memory or hold more objects than they really need. They are local classes to the corresponding Processor class, that does all the non-audio stuff.
    
!!! warning
    Keep in mind, that all the ACore classes run in a different thread, called by the sounddevice callback function.
    Therfore reading or writing to variables, except for initialization, has to be thread safe!
"""
from typing import TYPE_CHECKING, Optional, Tuple
from abc import ABC, abstractmethod
import numpy as np
import sounddevice as sd

if TYPE_CHECKING:
    from .types import InAs, AData, ABuf


class AGenerator(ABC):
    def __init__(self, out_buffer: bool, blocksize: int, start_frame: int) -> None:
        """This is the base class for audio generators.

        Args:
            out_buffer: Flag that decides if outputs are buffered.
            blocksize: Blocksize of the audio arrays.
            start_frame: The number of the first frame (to start counting from).
        """
        self._out_buffer = out_buffer
        self._blocksize = blocksize
        self._frame = start_frame

        # set output channels and update _out_buf
        self.out_chs = 1
        self._reload = True

    @property
    def out_chs(self) -> Optional[int]:
        return self._out_chs
    @out_chs.setter
    def out_chs(self, value: Optional[int]) -> None:
        self._out_chs = value
        # update _out_buf size
        if self._out_buffer and value is not None:
            self._out_buf = np.empty((self._blocksize, value), dtype=np.float32)
        else:
            self._out_buf = None

    def upstream(self, outdata: "AData", ch: int, frame: int) -> None:
        """This method is called by other AProcessors, connected to the outputs, to obtain the outputs data of the given channel.
        It is called in the opposite of audio flow and is therefore called upstream.
        The connected (other) AProcessors pass their outdata reference (to write to) and the channel ch they want to obtain the data from.
        Inside upstream, buffering and the appropriate calls for _mod and _inc are handled.

        Args:
            outdata: Reference of the 1D array to write to, with length blocksize.
            ch: The channel corresponding to the outdata at hand (zero indexed).
            frame: Current frame number to be processed, this always increases by +1 and is incremented by the callback.
        """
        # if the next frame is called, increment and set buffer reload flag
        if frame != self._frame:
            self._frame = frame
            self._inc()
            self._reload = True
        # if out_buffer is enabled and reload flag is set fill _out_buf
        if self._out_buffer and self._reload:
            if self._out_buffer:
                for out_ch in range(self._out_chs):
                    self._mod(self._out_buf[:, out_ch], out_ch)
        self._reload = False

        # if buffer is enabled return buffer
        if self._out_buffer:
            np.copyto(outdata, self._out_buf[:, ch], casting="no")
        # otherwise process given array
        else:
            self._mod(outdata, ch)

    @abstractmethod
    def _mod(self, outdata: "AData", ch: int) -> None:
        """This function is envoked by `upstream()`. 
        It should write something in outdata for the given output channel ch.
        Make sure to copy your data or write directly into outdata and not just set outdata to a new reference.
        See sounddevice callback manual for more details.

        Args:
            outdata: The 1D array to write to, with length blocksize.
            ch: The channel corresponding to the outdata at hand (zero indexed).
        """
    
    @abstractmethod
    def _inc(self) -> None:
        """This function is envoked by `upstream()`. 
        If the class changes over time, this function can be used to perform these changes.
        It is called exactly once after all channels of the class have been processed.
        """


class AEffect(ABC):
    def __init__(self, in_buffer: bool, out_buffer: bool, blocksize: int, start_frame: int) -> None:
        """This is the base class for audio effects

        Args:
            in_buffer: Flag that decides if inputs are buffered.
            out_buffer: Flag that decides if outputs are buffered.
            blocksize: Blocksize of the audio arrays.
            start_frame: The number of the first frame (to start counting from).
        """
        self._in_buffer = in_buffer
        self._out_buffer = out_buffer
        self._blocksize = blocksize
        self._frame = start_frame

        # set in-/output channels and update buffers
        self.in_as = None
        self.out_chs = None
        self._reload = True

    @property
    def in_as(self) -> "InAs":
        """A tupel defining what objects output and channel, the inputchannels are connected to.
        Evaluating in_as[in_ch] for an input channel in_ch of self, yields a tuple of (Connected Object, Connected Channel)."""
        return self._in_as
    @in_as.setter
    def in_as(self, value: "InAs") -> None:
        """Setting in_as automatically updates the buffer size, if in_buffer is enabled."""
        self._in_as = value
        # update _in_buf size
        if self._in_buffer and value is not None:
            self._in_buf = np.empty((self._blocksize, len(value)), dtype=np.float32)
        else:
            self._in_buf = None

    @property
    def out_chs(self) -> Optional[int]:
        return self._out_chs
    @out_chs.setter
    def out_chs(self, value: Optional[int]) -> None:
        self._out_chs = value
        # update _out_buf size
        if self._out_buffer and value is not None:
            self._out_buf = np.empty((self._blocksize, value), dtype=np.float32)
        else:
            self._out_buf = None

    def upstream(self, outdata: "AData", ch: int, frame: int) -> None:
        """This method is called by other AProcessors, connected to the outputs, to obtain the outputs data of the given channel.
        It is called in the opposite of audio flow and is therefore called upstream.
        The connected (other) AProcessors pass their outdata reference (to write to) and the channel ch they want to obtain the data from.
        Inside upstream, buffering and the appropriate calls for _mod and _inc are handled.

        Args:
            outdata: Reference of the 1D array to write to, with length blocksize.
            ch: The channel corresponding to the outdata at hand (zero indexed).
            frame: Current frame number to be processed, this always increases by +1 and is incremented by the callback.
        """
        # if the next frame is called, increment and set buffer reload flag
        if frame != self._frame:
            self._frame = frame
            self._inc()
            self._reload = True
        # if in_buffer is enabled and reload flag is set fill _in_buf
        if self._in_buffer and self._reload:
            self._in_buf.fill(0)
            for in_ch, in_a in enumerate(self._in_as):
                if in_a[0] is not None:
                    # send _in_buf upstream
                    in_a[0].upstream(self._in_buf[:, in_ch], in_a[1], frame)
        # if out_buffer is enabled and reload flag is set fill _out_buf
        if self._out_buffer and self._reload:
            for out_ch in range(self._out_chs):
                self._mod(self._out_buf[:, out_ch], out_ch)
        self._reload = False

        # if out_buffer is enabled copy _out_buf
        if self._out_buffer:
            np.copyto(outdata, self._out_buf[:, ch], casting="no")
        # otherwise process given array
        else:
            # _mod HAS TO HANDLE INPUT BUFFER + SETTING!!!
            self._mod(outdata, ch)

    @abstractmethod
    def _mod(self, outdata: "AData", ch: int) -> None:
        """This function is envoked by `upstream()`. 
        It should write something in outdata for the given output channel ch.
        Make sure to copy your data or write directly into outdata and not just set outdata to a new reference.
        See sounddevice callback manual for more details.

        Args:
            outdata: The 1D array to write to, with length blocksize.
            ch: The channel corresponding to the outdata at hand (zero indexed).

        Notes:
            The implementation of this function has to handle the `self.in_buffer` setting 
            and either process the input buffer `self.in_buf`, that is periodically filled by `start_upstream()`
            or obtain data from upstream and process it afterwards for each input channel.
            
        Example: Input buffer example
            ```python
            def _mod(self, outdata, ch):
                if self.in_buffer:
                    # process the self._in_buf[:, ch]
                    ...
                else:
                    if self._in_as[ch][0] is not None:
                        # get outdata from upstream
                        self._in_as[ch][0].upstream(outdata, self._in_as[ch][1], frame) 
                        # process outdata here
                        ...
            ```
        """
    
    @abstractmethod
    def _inc(self):
        """This function is envoked by `upstream()`. 
        If the class changes over time, this function can be used to perform these changes.
        It is called exactly once per audio frame, but not for the first one.
        """


class AAnalyzer(ABC):
    def __init__(self, in_buffer: bool, blocksize: int, start_frame: int) -> None:
        """This is the base class for audio analyzers.

        Args:
            in_buffer: Flag that decides if inputs are buffered.
            blocksize: Blocksize of the audio arrays.
            start_frame: The number of the first frame (to start counting from).
        """
        self._in_buffer = in_buffer
        self._blocksize = blocksize
        self._frame = start_frame

        # set input channels and update _in_buf
        self.in_as = None

    @property
    def in_as(self) -> "InAs":
        """A tupel defining what objects output and channel, the inputchannels are connected to.
        Evaluating in_as[in_ch] for an input channel in_ch of self, yields a tuple of (Connected Object, Connected Channel)."""
        return self._in_as
    @in_as.setter
    def in_as(self, value: "InAs"):
        """Setting in_as automatically updates the buffer size, if in_buffer is enabled."""
        self._in_as = value
        # update _in_buf size
        if self._in_buffer and value is not None:
            self._in_buf = np.empty((self._blocksize, len(value)), dtype=np.float32)
        else:
            self._in_buf = None

    def start_upstream(self, frame: int) -> None:
        # if in_buffer is enabled fill _in_buf
        if self._in_buffer:
            self._in_buf.fill(0)
            for in_ch, in_a in enumerate(self._in_as):
                if in_a[0] is not None:
                    # send _in_buf upstream
                    in_a[0].upstream(self._in_buf[:, in_ch], in_a[1], frame)

        # _process HAS TO HANDLE INPUT BUFFER + SETTING!!!
        self._process()

    @abstractmethod
    def _process(self) -> None:
        """This method is called once per audio frame to start the upstream chain.
        It either processes the obtained input buffer `self._in_buf`or sends data upstream.

        Notes:
            The implementation of this function has to handle the `self._in_buffer` setting 
            and either process the input buffer `self._in_buf`, that is periodically filled by `start_upstream()`
            or send data upstream to process for each input channel.

        Example: Input buffer example
            ```python
            if self._in_buffer:
                # process the self._in_buf
            else:
                for in_ch, in_a in enumerate(self._in_as):
                    if in_a[0] is not None:
                        in_a[0].upstream(NUMPY_ARRAY_TO_WRITE_TO[:, in_ch], in_a[1], frame)
                    else:
                        NUMPY_ARRAY_TO_WRITE_TO[:, in_ch].fill(0)
            ```
        """


class AInterface:
    def __init__(self, blocksize: int, start_frame: int):
        """This is the base class of the audio interface.
        It is used to assemble the callback function.

        Args:
            blocksize: Blocksize of the audio arrays.
            start_frame: The number of the first frame (to start counting from).
        """
        self.in_as = None
        self.alzs = None
        self.out_chs = None

        self._blocksize = blocksize
        self._frame = start_frame

        self.end_frame = None

    @property
    def in_as(self) -> "InAs":
        """A tupel defining what objects output and channel, the inputchannels are connected to.
        Evaluating in_as[in_ch] for an input channel in_ch of self, yields a tuple of (Connected Object, Connected Channel)."""
        return self._in_as
    @in_as.setter
    def in_as(self, value: "InAs"):
        self._in_as = value

    @property
    def out_chs(self) -> Optional[int]:
        return self._out_chs
    @out_chs.setter
    def out_chs(self, value: Optional[int]) -> None:
        self._out_chs = value
        if value is not None and value > 0:
            # update _out_buf size
            self._out_buf = np.empty((self._blocksize, value), dtype=np.float32)
        else:
            self._out_buf = None

    @property
    def alzs(self) -> Tuple["AAnalyzer"]:
        return self._alzs
    @alzs.setter
    def alzs(self, value: Tuple["AAnalyzer"]):
        self._alzs = value

    def callback(self, indata: np.ndarray, outdata: np.ndarray, frames: int, ctime, status):
        outdata.fill(0)
        # copy indata so it can be processed by upstream()
        if self._out_buf is not None:
            np.copyto(self._out_buf, indata)
        # call upstream method of the outputs connected to the inputs
        if self.in_as is not None:
            for in_ch, in_a in enumerate(self._in_as):
                if in_a[0] is not None:
                    in_a[0].upstream(outdata[:, in_ch], in_a[1], self._frame)
        # call AAnalyzers start_upstream method (because they wont get called otherwise)
        for alz in self._alzs:
            alz.start_upstream(self._frame)
        self._frame += 1 # Overflow?
        if self.end_frame is not None and self.end_frame <= self._frame:
            raise sd.CallbackStop

    def upstream(self, outdata: "AData", ch: int, frame: int):
        """This method is called by other AProcessors, connected to the outputs, to obtain the outputs data of the given channel.
        It just copies the bufferd indata of the respected channel to the given outdata reference

        Args:
            outdata: Reference of the 1D array to write to, with length blocksize.
            ch: The channel corresponding to the outdata at hand (zero indexed).
            frame: Current frame number to be processed, this always increases by +1 and is incremented by the callback.
        """
        np.copyto(outdata, self._out_buf[:, ch])



"""test_afile.py
This tests the serialization and deserialization of Interface properties to and from an AFile.
This test only works for `no_audio_device = True`, because the GitLab pipeline has no audio devices."""
import os
import asmu

def test_asetup():
    # create setup
    asetup = asmu.ASetup("./test_asetup_1.asmu")

    # create interface with all defaults,
    # register asetup and save
    interface1 = asmu.Interface(analog_input_channels=[1,2,3,4],
                                analog_output_channels=[1,2,3,4,5],
                                no_audio_device = True)
    asetup.interface = interface1
    asetup.save()

    # create other interface without paramters,
    # load setiings from asetup (of other interface),
    # and save as different filename
    interface2 = asmu.Interface(asetup = asetup,
                                device = None,
                                samplerate = None,
                                blocksize = None,
                                no_audio_device = True)
    asetup.save(path = "./test_asetup_2.asmu")

    # create another interface with different parameters,
    # register asetup and save
    interface3 = asmu.Interface(samplerate = 96000,
                                blocksize = 256,
                                analog_input_channels=[6,7],
                                analog_output_channels=[6,7,8],
                                no_audio_device = True)
    asetup.interface = interface3
    asetup.save(path = "./test_asetup_3.asmu")

    # compare the files
    with open("./test_asetup_1.asmu", "r", encoding="utf-8") as f1, \
        open("./test_asetup_2.asmu", "r", encoding="utf-8") as f2, \
        open("./test_asetup_3.asmu", "r", encoding="utf-8") as f3:
        content_1 = f1.read()
        content_2 = f2.read()
        content_3 = f3.read()
        assert content_1 == content_2, "Files content do not match!"
        assert content_1 != content_3, "Files content match, allthough they shouldn't!"

    # remove files
    os.remove("./test_asetup_1.asmu")
    os.remove("./test_asetup_2.asmu")
    os.remove("./test_asetup_3.asmu")

if __name__ == "__main__":
    test_asetup()



"""This example can be used to calibrate an Interface IInput and IOutput channel for pressure."""
from math import sqrt
import asmu

def calibrate_iinput_cPa(interface: "asmu.Interface", in_ch: int):
    calcPa = asmu.analyzer.CalIInput(interface, 94, "SPL", gain = 30, averages=100)
    interface.iinput(ch = in_ch).connect(calcPa.input())

    stream = interface.start()
    while not calcPa.finished():
        pass
    stream.stop()
    stream.close()
    print(calcPa.evaluate())
    print(f"cPa = {interface.iinput(ch = in_ch).cPa}")
    print(f"fPa = {interface.iinput(ch = in_ch).fPa}")

def calibrate_ioutput_cPa(interface: "asmu.Interface", in_ch: int, out_ch: int):
    outgain = 0.01
    sine = asmu.generator.Sine(interface, 1000)
    gain = asmu.effect.Gain(interface, outgain)
    calcPa = asmu.analyzer.CalIOutput(interface, outgain, "Pa", interface.ioutput(ch = out_ch), averages=100)
    sine.output().connect(gain.input())
    gain.output().connect(interface.ioutput(ch = out_ch))
    interface.iinput(ch = in_ch).connect(calcPa.input())

    stream = interface.start()
    while not calcPa.finished():
        pass
    stream.stop()
    stream.close()
    print(calcPa.evaluate())
    print(f"cPa = {interface.ioutput(ch = out_ch).cPa}")
    print(f"fPa = {interface.ioutput(ch = out_ch).fPa}")

if __name__ == "__main__":
    in_ch = 9
    out_ch = 3
    interface = asmu.Interface(device = "ASIO MADIface USB",
                               analog_input_channels=[in_ch],
                               analog_output_channels=[out_ch],
                               blocksize=8192,
                               samplerate=192000)

    print(f"Connect microphone in 94dB calibrator to input channel {in_ch}.")
    input("\tPress ENTER to start.")
    calibrate_iinput_cPa(interface, in_ch)

    print(f"Remove the microphone from the calibrater and place it next to a source driven by to the output channel {out_ch}?")
    input("\tPress ENTER to start.")
    calibrate_ioutput_cPa(interface, in_ch, out_ch)

    if input("Start generator? (y|n)") == "y":
        # verify if everything worked correctly
        sine = asmu.generator.Sine(interface, 1000)
        SPL = 60 # dB
        Pap = 2e-5 * 10**(SPL/20) * sqrt(2) # set desired peak amplitude
        gain = asmu.effect.Gain(interface, Pap/interface.ioutput().cPa)
        sine.output().connect(gain.input())
        gain.output().connect(interface.ioutput(ch = out_ch))
        print("Starting sine generator...")
        stream = interface.start()
        print(f"You now should measure a {SPL:.2f}dB sine wave on the output channel {interface.ioutput().channel}.")
        input("\tPress ENTER to stop.")
        stream.stop()
        stream.close()




"""Groove out to the beats and rythms of the legendary amen break and a funky baseline.
Try to sing along, your masterpiece is recorded under `./recordings/amen_recording.wav`
When you are finished press Strg + C to stop and save the recording.
HAVE FUN!"""
import asmu

def main():
    # create interface
    interface = asmu.Interface(samplerate=44100,
                               blocksize=256,
                               analog_input_channels=[1],
                               analog_output_channels=[1, 2])
    # open files
    with asmu.AFile(interface, mode = "r", path="./recordings/amen.wav", channels = 2) as drum_afile,\
        asmu.AFile(interface, mode = "w+", path="./recordings/amen_recording.wav", channels = 2) as rec_afile:
        # create objects
        bass = asmu.generator.Sine(interface, 98)
        drums = asmu.generator.Player(interface, drum_afile, loop = True)
        sumL = asmu.effect.Sum(interface)
        sumR = asmu.effect.Sum(interface)
        bassRamp = asmu.effect.GainRamp(interface, 1, 0.1, scale="log", in_buffer=False)
        rec = asmu.analyzer.Recorder(interface, rec_afile)

        # establish connections
        bass.output().connect(bassRamp.input(0))

        drums.output(0).connect(sumL.input(0))
        drums.output(1).connect(sumR.input(0))
        bassRamp.output(0).connect(sumL.input(1))
        bassRamp.output(0).connect(sumR.input(1))
        interface.iinput(ch=1).connect(sumL.input(2))
        interface.iinput(ch=1).connect(sumR.input(2))

        sumL.output().connect(interface.ioutput(ch=1))
        sumR.output().connect(interface.ioutput(ch=2))

        sumL.output().connect(rec.input(0))
        sumR.output().connect(rec.input(1))

        # start audio for precisely N frames
        stream = interface.start(end_frame = 1000)

        # sequencer
        try:
            bar = 1
            while stream.active:
                if drums.looped(timeout = 1):
                    print(f"looped bar= {bar}")
                    match bar:
                        case 0:
                            bassRamp.set_gain(1.0)
                            bass.set_frequency(98) # G2
                        case 1:
                            bass.set_frequency(65.4) # C2
                        case 2:
                            bass.set_frequency(49) # G1
                        case 3:
                            bassRamp.set_gain(0.5)
                    bar += 1
                    bar %= 4
        except KeyboardInterrupt:
            stream.stop()
        stream.close()

if __name__ == "__main__":
    main()










import logging
from typing import TYPE_CHECKING, List, Tuple, Union, Optional, Callable
import numpy as np
import sounddevice as sd

from .acore import AInterface
from .processor import Processor
from .io import IInput, IOutput

if TYPE_CHECKING:
    from .asetup import ASetup
    from .processor import AAnalyzer

logger = logging.getLogger(__name__)

START_FRAME = 0


class Interface(Processor):
    def __init__(self, 
                 asetup: Optional["ASetup"] = None,
                 device: Optional[Union[Tuple[Union[str, int], Union[str, int]], Union[str, int]]] = None,
                 samplerate: Optional[int] = 44100,
                 blocksize: Optional[int] = 1024,
                 analog_input_channels: List[int] = None,
                 analog_output_channels: List[int] = None,
                 no_audio_device: bool = False):
        """The Interface class represents the audio interface or soundcard. 
        It is holding the audio generator and manages settings.
        The settings can either be specified on intialization, by an ASetup class, or used as default.

        Args:
            asetup: Reference to an ASetup instance. If set, loads the settings from the given ASetup.
                If you dont want that, specify it after initialization, by setting `Interface.asetup = ASetup`.
            device: Device name, or tuple of device names for different input and output device. If None, the default device(s) are used.
            samplerate: The samplerate in samples per second.
            blocksize: The blocksize defines the samples per frame.
            analog_input_channels: List if analog input channels, only works for ASIO, CoreAudio devices, or no_audio_device.
            analog_output_channels: List if analog output channels, only works for ASIO, CoreAudio devices, or no_audio_device.
            no_audio_device: Flag used for testing. Skips audio device and allows the direct use of the callback function.

        Notes:
            The device name can be optained by running the following code snippet.
            ```python linenums="1" title="List audio devices"
            import asmu
            asmu.query_devices()
            ```
        """
        # this is used for the analyzers to add themselfes later
        self._analyzers: Tuple["AAnalyzer"] = ()

        self.asetup = asetup
        if asetup is not None:
            asetup.load()
        else:
            # init from given values
            self._samplerate = samplerate
            self._blocksize = blocksize
            self.latency = 0
            # init device
            self._device = device
            if device is None:
                # if None use default
                self._device = sd.default.device
            elif not isinstance(self._device, tuple):
                # if single device use for input and output
                self._device = (self._device, self._device)
            self._no_audio_device = no_audio_device
            if not no_audio_device:
                if not self._is_asio() and not self._is_ca():
                    # if not asio or coraaudio overwrite the channel lists
                    logger.warning("No ASIO or CoreAudio device specified, channel numbers are ignored.")
                    analog_input_channels = list(range(sd.query_devices(device=self._device[0], kind="input")["max_input_channels"]))
                    analog_input_channels = [ch+1 for ch in analog_input_channels]
                    analog_output_channels = list(range(sd.query_devices(device=self._device[1], kind="output")["max_output_channels"]))
                    analog_output_channels = [ch+1 for ch in analog_output_channels]
            # set in-/outputs accordingly
            self._iinputs = ()
            if self._device[0] is not None and analog_input_channels is not None:
                self._iinputs = tuple(IInput(self, iin_ch) for iin_ch in analog_input_channels)
            self._ioutputs = ()
            if self._device[1] is not None and analog_output_channels is not None:
                self._ioutputs = tuple(IOutput(self, iout_ch) for iout_ch in analog_output_channels)

            self._ainterface = AInterface(blocksize = self._blocksize,
                                        start_frame = START_FRAME)
        super().__init__(self)

    def __del__(self):
        # deregister from asetup
        if self._asetup is not None:
            self._asetup.interface = None

    @property
    def samplerate(self):
        return self._samplerate

    @property
    def blocksize(self):
        return self._blocksize

    @property
    def device(self):
        return self._device

    @property
    def start_frame(self) -> int:
        return START_FRAME
    
    @property
    def callback(self) -> Callable:
        return self._ainterface.callback
    
    @property
    def asetup(self) -> Optional["ASetup"]:
        return self._asetup
    @asetup.setter
    def asetup(self, value: Optional["ASetup"]):
        self._asetup = value
        # register in asetup
        if value is not None:
            self._asetup.interface = self

    @property
    def analyzers(self):
        return self._analyzers
    @analyzers.setter
    def analyzers(self, value):
        self._analyzers = value
        self.update_acore()

    @property
    def acore(self):
        return self._ainterface

    @property
    def outputs(self) -> Tuple["IInput"]:
        return self._iinputs

    def iinput(self, idx: int=0, ch: Optional[int] = None, name: Optional[str] = None) -> "IInput":
        """Get the interface analog IInput, which is an asmu Output, by the given argument.

        Args:
            idx (int, optional): Index in zero indexed list of IInputs.
            ch (int, optional): Interface analog input channel, stored in the IInput.
            name (str, optional): Given name of the channel, stored in the IInput.

        Raises:
            ValueError: No IInput for the given argument registered

        Returns:
            Reference to IInput object.
        """
        if ch is not None:
            try:
                return next((outpu for outpu in self._iinputs if outpu.channel == ch))
            except StopIteration:
                raise ValueError(f"No IInput on channel {ch} registered.")
        if name is not None:
            try:
                return next((outpu for outpu in self._iinputs if outpu.name == name))
            except StopIteration:
                raise ValueError(f"No IInput with name {name} registered.")
        return self._iinputs[idx]

    def ioutput(self, idx: int=0, ch: Optional[int] = None, name: Optional[str] = None) -> "IOutput":
        """Get the interface analog IOutput, which is an asmu Input, by the given argument.

        Args:
            idx (int, optional): Index in zero indexed list of IOutputs.
            ch (int, optional): Interface analog output channel, stored in the IOutput.
            name (str, optional): Given name of the channel, stored in the IOutput.

        Raises:
            ValueError: No IOutput for the given argument registered

        Returns:
            Reference to IOutput object.
        """
        if ch is not None:
            try:
                return next((inpu for inpu in self._ioutputs if inpu.channel == ch))
            except StopIteration:
                raise ValueError(f"No IOutput on channel {ch} registered.")
        if name is not None:
            try:
                return next((inpu for inpu in self._ioutputs if inpu.name == name))
            except StopIteration:
                raise ValueError(f"No IOutput with name {name} registered.")
        return self._ioutputs[idx]

    def _cal_latency(self, time):
        """Calculate and store loopback latency without physical connection.

        !!! warning
            Dont rely on this method, as it only calculates the ADC/DAC's internal latency. 
            Use [latency_from_rec.py](../examples.md/#latency_from_rec.py) to compare this result with the real loopback calibration.

        Args:
            time (CData): The time object given in the callback function.
        """
        self.latency = round((time.outputBufferDacTime-time.inputBufferAdcTime)*self.samplerate + 1) # the +1 was measured experimentally (could be the cable?)

    def _is_asio(self) -> bool:
        """Determine if ALL of the set io devices are ASIO compatible.

        Returns:
            `True`, when all given devices are ASIO compatible. `False` otherwise.
        """
        asio = True
        if self._device[0] is not None:
            asio = ("asio" in str(sd.query_devices(device = self._device[0], kind = "input")["name"]).lower())
        if self._device[1] is not None:
            asio = ("asio" in str(sd.query_devices(device = self._device[1], kind = "output")["name"]).lower())
        return asio
    
    def _is_ca(self) -> bool:
        """Determine if ALL of the set io devices are CoraAudio compatible.

        Returns:
            `True`, when all given devices are CoraAudio compatible. `False` otherwise.
        """
        ca = True
        if self._device[0] is not None:
            ca = ("coreaudio" in str(sd.query_devices(device = self._device[0], kind = "input")["name"]).lower())
        if self._device[1] is not None:
            ca = ("coreaudio" in str(sd.query_devices(device = self._device[1], kind = "output")["name"]).lower())
        return ca
    
    def _is_test(self) -> bool:
        """Determine if at least one of the set io devices is set to test.

        Returns:
            `True`, when at least one given devices is set to test. `False` otherwise.
        """
        test = False
        if self._device[0] is not None:
            test = ("test" in str(self._device[0]).lower())
        if self._device[1] is not None:
            test = ("test" in str(self._device[1]).lower())
        return test

    # SOUNDDEVICE 
    def _init_sounddevice(self) -> None:
        """Initiializes sounddevice with the classes attributes for the given lists of inputs and outputs.
        """
        stream = sd.default
        stream.dtype = np.float32
        if self.samplerate is not None:
            stream.samplerate = self.samplerate
        if self.blocksize is not None:
            stream.blocksize = self.blocksize
        stream.device = self.device
        if self._is_asio():
            if self._iinputs:
                in_channels = [inpu.channel - 1 for inpu in self._iinputs] # convert to channel names starting with 0
                asio_in = sd.AsioSettings(channel_selectors=in_channels)

                if not self._ioutputs:
                    stream.extra_settings = asio_in
                    stream.channels = len(in_channels)
                    return

            if self._ioutputs:
                out_channels = [output.channel - 1 for output in self._ioutputs]
                asio_out = sd.AsioSettings(channel_selectors=out_channels)

                if not self._iinputs:
                    stream.extra_settings = asio_out
                    stream.channels = len(out_channels)
                    return
            
            if self._iinputs and self._ioutputs:
                stream.extra_settings = (asio_in, asio_out)
                stream.channels = (len(in_channels), len(out_channels))
                return
            
        elif self._is_ca(): 
            raise NotImplementedError("CoreAudio channel selection is not tested!")
            if self._iinputs:
                in_channels = [inpu.channel - 1 for inpu in self._iinputs] # convert to channel names starting with 0
                ca_in = sd.CoreAudioSettings(channel_map=in_channels)

                if not self._ioutputs:
                    stream.extra_settings = ca_in
                    stream.channels = len(in_channels)
                    return
                
            if self._ioutputs:
                out_channels = [-1]*sd.query_devices(device=self.device, kind="output")["max_output_channels"]
                for idx, c in enumerate(self._ioutputs):
                    out_channels[c.channel -1] = idx
                ca_out = sd.CoreAudioSettings(channel_map=out_channels)

                if not self._iinputs:
                    stream.extra_settings = ca_out
                    stream.channels = len(out_channels)
                    return

            if self._iinputs and self._ioutputs:
                stream.extra_settings = (ca_in, ca_out)
                stream.channels = (len(in_channels), len(out_channels))
                return

    def start(self, end_frame: int = None) -> sd.Stream:
        """Start the audio stream.

        Args:
            end_frame: If set, the stream is stopped at the given end_frame.

        Returns:
            Reference to the started sounddevice stream. The full documentation is linked [here](https://python-sounddevice.readthedocs.io/en/latest/api/streams.html#sounddevice.Stream).
                But the basic functions can be summerized as:

                - `Stream.active`  : `True`, when the stream is active. `False` otherwise. This is useful when end_frame is used, to check if the stream is finished. 
                - `Stream.stop()`  : Terminate audio processing. This waits until all pending audio buffers have been played before it returns.
                - `Stream.close()` : Close the stream. This should be used after the stream has been stopped because the end_frame has been reached or Stream.stop has been called.

                If it is used for an active stream, the audio buffers are discarded.
        """
        if self._no_audio_device:
            raise ValueError("Starting a stream in no_audio_device mode is not possible!")
        self._init_sounddevice()
        self._ainterface.end_frame = end_frame
        stream = sd.Stream(callback=self._ainterface.callback)
        stream.start()
        return stream
        
    def update_acore(self):
        # create in_as tuple
        in_as = ()
        for inp in self._ioutputs:
            # add proper connection constraint
            if inp.connected_output is None:
                in_as += ((None, 0), )
            else:
                # find channel idx it is connected to
                in_as += ((inp.connected_output.acore, inp.connected_output.idx), )
        self._ainterface.in_as = in_as
        # count outputs that have a connection
        self._ainterface.out_chs = len(self._iinputs)
        # update alz
        self._ainterface.alzs = tuple(alz.acore for alz in self.analyzers)

    def serialize(self):
        data = {}
        data["samplerate"] = int(self._samplerate)
        data["blocksize"] = int(self._blocksize)
        data["latency"] = int(self.latency)
        data["no_audio_device"] = bool(self._no_audio_device)

        if not self._no_audio_device:
            device = ()
            if self._device[0] is not None:
                device += (sd.query_devices(device = self._device[0], kind = "input"), )
            else:
                device += (None, )
            if self._device[1] is not None:
                device += (sd.query_devices(device = self._device[1], kind = "output"), )
            else:
                device += (None, )
        data["device"] = tuple(self._device)

        iinputs = []
        for iinput in self._iinputs:
            iinputs.append(iinput.serialize())
        data["iinputs"] = iinputs

        ioutputs = []
        for ioutput in self._ioutputs:
            ioutputs.append(ioutput.serialize())
        data["ioutputs"] = ioutputs
        return data

    def deserialize(self, data: dict):
        self._samplerate = int(data["samplerate"])
        self._blocksize = int(data["blocksize"])
        self.latency = int(data["latency"])
        self._no_audio_device = bool(data["no_audio_device"])
        
        self._device = data["device"]

        self._iinputs = ()
        for iinput_data in data["iinputs"]:
            iinput = IInput(self, iinput_data["channel"])
            iinput.deserialize(iinput_data)
            self._iinputs += (iinput, )

        self._ioutputs = ()
        for ioutput_data in data["ioutputs"]:
            ioutput = IOutput(self, ioutput_data["channel"])
            ioutput.deserialize(ioutput_data)
            self._ioutputs += (ioutput, )

        self._ainterface = AInterface(blocksize = self._blocksize,
                                      start_frame = START_FRAME)



import tempfile
import pathlib
import logging
import json
from typing import TYPE_CHECKING, Literal
from datetime import datetime
import soundfile
import numpy as np

if TYPE_CHECKING:
    from .interface import Interface
    from .types import ABuf

logger = logging.getLogger(__name__)


class AFile(soundfile.SoundFile):
    def __init__(self, interface: "Interface", mode: Literal["r", "w", "w+"] = "r", path:str=None, channels:int=None, temp:bool=False) -> None:
        """The AFile class handles sound files. Metadata is used to store and load additional settings.
        It is recommended to open AFile's in a `with` context manager. 
        At the moment only .wav files with 24-bit encoding are supported.

        Args:
            interface: Reference to an Interface instance.
            mode: The mode to open the file with..
            path: The filepath.
            channels: Number of channels.
            temp: When true, the file is temporary and deleted on close, otherwise the file is persistent.

        Raises:
            ValueError: For incorrect input arguments
            FileNotFoundError: When the path does not exist.

        Example: Find shape of an audio file
            ```python
            import asmu
            with asmu.AFile(interface, mode = "r", path="myfile.wav") as afile:
                print(afile.data.shape)
            ```
        """
        # Check input
        if not mode in ["r", "w", "w+"]: raise ValueError("Specify a valid mode of [\"r\", \"w\", \"w+\"].")
        if path is None and temp is False: raise ValueError("For non-temporary files, a path has to be specified")
        if mode != "r" and channels is None: raise ValueError("Either read (mode = \"r\") a file or specify channels.")

        self._interface = interface
        self._mode = mode
        self.path = path
        if path is not None: 
            self.path = pathlib.Path(path)
            if not self.path.exists() and mode == "r": raise FileNotFoundError(f"File {path} does not exist.")
            if self.path.exists(): logger.debug(f"File {path} exists.")
        self._channels = channels
        self.temp = temp
        self._settings = {}
    
    # store settings as json string in metadata "comment"
    @property
    def settings(self) -> dict:
        """Additional JSON settings in the metadata's comment field."""
        return self._settings
    @settings.setter
    def settings(self, value: dict) -> None:
        self._settings = value
        self.comment = json.dumps(self.settings)

    @property
    def data(self) -> "ABuf":
        """This property can be used to access the data of the AFile.
        Dont use it during an active audio stream, as it resets the file pointer to the
        start of the file.

        Returns:
            Data of AFile as numpy array of shape (Samples x Channels).
        """
        self.flush()
        self.seek(0)
        return self.read(dtype="float32", always_2d=True)
    
    def __enter__(self):
        return self.open()
    
    def __exit__(self, *args):
        self.close()
        return False

    def open(self) -> "AFile":
        """Open a file, with respect to the settings specified at initialization.

        Returns:
            An instance of th opened AFile.
        """
        if self._mode == "r":
            super().__init__(self.path, mode=self._mode)
            # load settings
            if self.comment:
                self._settings = json.loads(self.comment)
        else:
            if self.temp:
                if self.path is None:
                    tmp = tempfile.NamedTemporaryFile(suffix='.wav', delete=True)
                else:
                    tmp = tempfile.NamedTemporaryFile(prefix=self.path.stem, suffix='.wav', dir=self.path.parent, delete=True)
                super().__init__(tmp, mode=self._mode, samplerate=self._interface.samplerate, channels=self._channels, subtype="PCM_24", format="WAV")
            else:
                super().__init__(self.path, mode=self._mode, samplerate=self._interface.samplerate, channels=self._channels, subtype="PCM_24", format="WAV")
            now = datetime.now()
            # set wav metadata
            self.title = self.path.stem
            self.date = now.strftime("%Y-%m-%dT%H:%M:%S%z") # ISO 8601
        return self


"""This submodule holds the base classes for Processors, that form the main building blocks auf an audio processing chain. 
These classes hold the underlying ACore object and hanled their initialization, connection, 
and `threading.Event()` or `queueing.Queue()` based communication with the main program during an active audio stream."""
import logging
from typing import TYPE_CHECKING, Tuple
from abc import ABC, abstractmethod

from .io import Input, Output

if TYPE_CHECKING:
    from .acore import AGenerator, AEffect, AAnalyzer
    from .interface import Interface
    from .types import ACore

logger = logging.getLogger(__name__)


class Processor(ABC):
    def __init__(self, interface: "Interface") -> None:
        self._interface = interface

    @property
    @abstractmethod
    def acore(self) -> "ACore":
        """Returns the ACore element of the processor."""

    @abstractmethod
    def update_acore(self) -> None:
        """Set in_as and out_chs of acore. This is called by the inputs/outputs."""


class Generator(Processor):
    def __init__(self, agenerator: "AGenerator", interface: "Interface", outputs: Tuple["Output"], out_update: bool) -> None:
        """This is the base class for generators, holding the audio generator.

        Args:
            agenerator: Reference to the corresponding ACore object.
            interface: Reference to an Interface instance.
            outputs: A tuple of Output instances.
            out_update: Flag that decides if dynamic output updates are enabled.
        """
        self._agenerator = agenerator
        self._outputs = outputs
        self._out_update = out_update
        # update_acore() is not called here, because it is called on the first connection anyways!
        super().__init__(interface)

    @property
    def acore(self) -> "AGenerator":
        return self._agenerator

    @property
    def outputs(self) -> Tuple["Output"]:
        return self._outputs

    def output(self, idx: int = 0) -> "Output":
        """Get the Output for the given index.
        For multiple output Generators that support output-update.
        when the index is +1 of the existing Output(s) a new Output is added.

        Args:
            idx: Index in zero indexed list of Outputs.

        Returns:
            Reference to Output object.
        """
        while idx > len(self._outputs) -1 and self._out_update:
            self._outputs += (Output(self), )
        return self._outputs[idx]

    def update_acore(self) -> None:
        self._agenerator.out_chs = len(self._outputs)


class Effect(Processor):
    def __init__(self, aeffect: "AEffect", interface: "Interface", inputs: Tuple["Input"], outputs: Tuple["Output"], in_update: bool, out_update: bool) -> None:
        """This is the base class for effects, holding the audio effect.

        Args:
            aeffect: Reference to the corresponding ACore object.
            interface: Reference to an Interface instance.
            inputs: A tuple of Input instances.
            outputs: A tuple of Output instances.
            in_update: Flag that decides if dynamic input updates are enabled.
            out_update: Flag that decides if dynamic output updates are enabled.

        Notes:
            If both update flags are enabled, input and output counts are syncronized.
        """
        self._aeffect = aeffect
        self._inputs = inputs
        self._outputs = outputs
        self._in_update = in_update
        self._out_update = out_update
        # update_acore() is not called here, because it is called on the first connection anyways!
        super().__init__(interface)

    @property
    def acore(self) -> "AEffect":
        return self._aeffect

    @property
    def outputs(self) -> Tuple["Output"]:
        return self._outputs

    def input(self, idx: int = 0) -> "Input":
        """Get the Input for the given index.
        For multiple input Effects that support input-update.
        when the index is +1 of the existing Input(s) a new Input is added.
        When output-update is also supported, Input(s) and Output(s) are syncronized.

        Args:
            idx: Index in zero indexed list of Inputs.

        Returns:
            Reference to Input object.
        """
        while idx > len(self._inputs) -1 and self._in_update:
            self._inputs += (Input(self), )
            # if update flags are true, keep channel count equal
            if self._out_update: self._outputs += (Output(self), )
        return self._inputs[idx]

    def output(self, idx: int = 0) -> "Output":
        """Get the Output for the given index.
        For multiple output Effects that support output-update.
        when the index is +1 of the existing Output(s) a new Output is added.
        When input-update is also supported, Input(s) and Output(s) are syncronized.

        Args:
            idx: Index in zero indexed list of Outputs.

        Returns:
            Reference to Output object.
        """
        while idx > len(self._outputs) -1 and self._out_update:
            self._outputs += (Output(self), )
            # if update flags are true, keep channel count equal
            if self._in_update: self._inputs += (Input(self), )
        return self._outputs[idx]

    def update_acore(self) -> None:
        # create in_as tuple
        in_as = ()
        for inp in self._inputs:
            # add proper connection constraint
            if inp.connected_output is None:
                in_as += ((None, 0), )
                logger.info(f"Input {inp} is not yet connected to any outputs.")
            else:
                # find channel idx it is connected to
                in_as += ((inp.connected_output.acore, inp.connected_output.idx), )
        self._aeffect.in_as = in_as
        # count outputs that have a connection
        self._aeffect.out_chs = len(self._outputs)


class Analyzer(Processor):
    def __init__(self, aanalyzer: "AAnalyzer", interface: "Interface", inputs: Tuple["Input"], in_update: bool) -> None:
        """This is the base class for analyzers, holding the audio analyzer.

        Args:
            aanalyzer: Reference to the corresponding ACore object.
            interface: Reference to an Interface instance.
            inputs: A tuple of Input instances.
            in_update: Flag that decides if dynamic input updates are enabled.
        """
        self._aanalyzer = aanalyzer
        self._inputs = inputs
        self._in_update = in_update
        # add to analyzer list in interface
        interface.analyzers += (self, )
        # update_acore() is not called here, because it is called on the first connection anyways!
        super().__init__(interface)

    @property
    def acore(self) -> "AAnalyzer":
        return self._aanalyzer

    def input(self, idx: int=0) -> "Input":
        """Get the Input for the given index.
        For multiple input Effects that support input-update.
        when the index is +1 of the existing Input(s) a new Input is added.

        Args:
            idx: Index in zero indexed list of Inputs.

        Returns:
            Reference to Input object.
        """
        while idx > len(self._inputs) -1 and self._in_update:
            self._inputs += (Input(self), )
        return self._inputs[idx]

    def update_acore(self) -> None:
        # create in_as tuple
        in_as = ()
        for inp in self._inputs:
            # add proper connection constraint
            if inp.connected_output is None:
                in_as += ((None, 0), )
                logger.info(f"Input {inp} is not yet connected to any outputs.")
            else:
                # find channel idx it is connected to
                in_as += ((inp.connected_output.acore, inp.connected_output.idx), )
        self._aanalyzer.in_as = in_as



"""This example calibrates the latency between given input and output."""
import time
import asmu
from asmu_utils.correlation import get_corrs_sampleshifts

def calibrate_latency(interface: "asmu.Interface", in_ch: int, out_ch: int) -> int:
    sineburst = asmu.generator.SineBurst(interface, 1000, 10)
    with asmu.AFile(interface, mode="w+", path="temp.wav", channels=2, temp=True) as afile:
        rec = asmu.analyzer.Recorder(interface, afile)

        sineburst.output().connect(interface.ioutput(ch = out_ch))
        sineburst.output().connect(rec.input(0))
        interface.iinput(ch = in_ch).connect(rec.input(1))

        stream = interface.start(end_frame=16)
        while stream.active:
            time.sleep(0.1)

        data = afile.data

    corrs, shifts = get_corrs_sampleshifts(data, data[:,0], 10/1000*interface.samplerate)
    assert shifts[0] == 0
    return shifts[1]

if __name__ == "__main__":
    in_ch = 3
    out_ch = 3
    interface = asmu.Interface(device = "ASIO MADIface USB",
                                analog_input_channels=[in_ch],
                                analog_output_channels=[out_ch],
                                blocksize=8192,
                                samplerate=192000)
    latency = calibrate_latency(interface, in_ch, out_ch)
    print(f"Found latency = {latency}")